@@ -25,6 +25,7 @@
 #include <gtk/gtk.h>
 
 #include <camel/camel.h>
+#include <libedataserver/libedataserver.h>
 
 #include <e-util/e-emoticon.h>
 #include <e-util/e-spell-checker.h>
@@ -32,6 +33,16 @@
 
 #define DEFAULT_CONTENT_EDITOR_NAME "WebKit"
 
+#define E_CONTENT_EDITOR_DIALOG_HRULE		"hrule"
+#define E_CONTENT_EDITOR_DIALOG_IMAGE		"image"
+#define E_CONTENT_EDITOR_DIALOG_LINK		"link"
+#define E_CONTENT_EDITOR_DIALOG_PAGE		"page"
+#define E_CONTENT_EDITOR_DIALOG_CELL		"cell"
+#define E_CONTENT_EDITOR_DIALOG_TABLE		"table"
+#define E_CONTENT_EDITOR_DIALOG_SPELLCHECK	"spellcheck"
+#define E_CONTENT_EDITOR_DIALOG_FIND		"find"
+#define E_CONTENT_EDITOR_DIALOG_REPLACE		"replace"
+
 G_BEGIN_DECLS
 
 struct _EHTMLEditor;
@@ -39,6 +50,8 @@ struct _EHTMLEditor;
 #define E_TYPE_CONTENT_EDITOR e_content_editor_get_type ()
 G_DECLARE_INTERFACE (EContentEditor, e_content_editor, E, CONTENT_EDITOR, GtkWidget)
 
+typedef GHashTable EContentEditorContentHash;
+
 typedef void (*EContentEditorInitializedCallback)	(EContentEditor *content_editor,
 							 gpointer user_data);
 
@@ -55,18 +68,20 @@ struct _EContentEditorInterface {
 							 const gchar *content,
 							 EContentEditorInsertContentFlags flags);
 
-	gchar *		(*get_content)			(EContentEditor *editor,
-							 EContentEditorGetContentFlags flags,
+	void		(*get_content)			(EContentEditor *editor,
+							 guint32 flags, /* bit-or of EContentEditorGetContentFlags */
 							 const gchar *inline_images_from_domain,
-							 GSList **inline_images_parts /* newly created CamelMimePart * */);
+							 GCancellable *cancellable,
+							 GAsyncReadyCallback callback,
+							 gpointer user_data);
+	EContentEditorContentHash *
+			(*get_content_finish)		(EContentEditor *editor,
+							 GAsyncResult *result,
+							 GError **error);
 
 	void		(*insert_image)			(EContentEditor *editor,
 							 const gchar *uri);
 
-	void            (*insert_image_from_mime_part)
-							(EContentEditor *editor,
-							 CamelMimePart *part);
-
 	void		(*insert_emoticon)		(EContentEditor *editor,
 							 EEmoticon *emoticon);
 
@@ -92,8 +107,6 @@ struct _EContentEditorInterface {
 	void		(*set_spell_checking_languages)	(EContentEditor *editor,
 							 const gchar **languages);
 
-	gchar *		(*get_selected_text)		(EContentEditor *editor);
-
 	gchar *		(*get_caret_word)		(EContentEditor *editor);
 
 	void		(*replace_caret_word)		(EContentEditor *editor,
@@ -105,9 +118,6 @@ struct _EContentEditorInterface {
 
 	void		(*selection_unindent)		(EContentEditor *editor);
 
-	void		(*selection_create_link)	(EContentEditor *editor,
-							 const gchar *uri);
-
 	void		(*selection_unlink)		(EContentEditor *editor);
 
 	void		(*find)				(EContentEditor *editor,
@@ -128,17 +138,14 @@ struct _EContentEditorInterface {
 
 	void		(*selection_wrap)		(EContentEditor *editor);
 
-	guint		(*get_caret_position)		(EContentEditor *editor);
-
-	guint		(*get_caret_offset)		(EContentEditor *editor);
-
 	gchar *		(*get_current_signature_uid)	(EContentEditor *editor);
 
 	gboolean	(*is_ready)			(EContentEditor *editor);
 
 	gchar *		(*insert_signature)		(EContentEditor *editor,
 							 const gchar *content,
-							 gboolean is_html,
+							 EContentEditorMode editor_mode,
+							 gboolean can_reposition_caret,
 							 const gchar *signature_id,
 							 gboolean *set_signature_from_message,
 							 gboolean *check_if_signature_is_changed,
@@ -160,9 +167,11 @@ struct _EContentEditorInterface {
 
 	void		(*insert_row_below)		(EContentEditor *editor);
 
-	gboolean	(*on_h_rule_dialog_open)	(EContentEditor *editor);
+	void		(*on_dialog_open)		(EContentEditor *editor,
+							 const gchar *name);
 
-	void		(*on_h_rule_dialog_close)	(EContentEditor *editor);
+	void		(*on_dialog_close)		(EContentEditor *editor,
+							 const gchar *name);
 
 	void		(*h_rule_set_align)		(EContentEditor *editor,
 							 const gchar *value);
@@ -186,10 +195,6 @@ struct _EContentEditorInterface {
 
 	gboolean	(*h_rule_get_no_shade)		(EContentEditor *editor);
 
-	void		(*on_image_dialog_open)		(EContentEditor *editor);
-
-	void		(*on_image_dialog_close)	(EContentEditor *editor);
-
 	void		(*image_set_src)		(EContentEditor *editor,
 							 const gchar *value);
 
@@ -245,21 +250,15 @@ struct _EContentEditorInterface {
 
 	gchar *		(*image_get_align)		(EContentEditor *editor);
 
-	void		(*on_link_dialog_open)		(EContentEditor *editor);
+	void		(*link_get_properties)		(EContentEditor *editor,
+							 gchar **out_href,
+							 gchar **out_text,
+							 gchar **out_name);
 
-	void		(*on_link_dialog_close)		(EContentEditor *editor);
-
-	void		(*link_get_values)		(EContentEditor *editor,
-							 gchar **href,
-							 gchar **text);
-
-	void		(*link_set_values)		(EContentEditor *editor,
+	void		(*link_set_properties)		(EContentEditor *editor,
 							 const gchar *href,
-							 const gchar *text);
-
-	void		(*on_page_dialog_open)		(EContentEditor *editor);
-
-	void		(*on_page_dialog_close)		(EContentEditor *editor);
+							 const gchar *text,
+							 const gchar *name);
 
 	void		(*page_set_text_color)		(EContentEditor *editor,
 							 const GdkRGBA *value);
@@ -284,6 +283,9 @@ struct _EContentEditorInterface {
 
 	void		(*page_get_visited_link_color)	(EContentEditor *editor,
 							 GdkRGBA *value);
+	void		(*page_set_font_name)		(EContentEditor *editor,
+							 const gchar *value);
+	const gchar *	(*page_get_font_name)		(EContentEditor *editor);
 
 	void		(*page_set_background_image_uri)
 							(EContentEditor *editor,
@@ -292,10 +294,6 @@ struct _EContentEditorInterface {
 	gchar *		(*page_get_background_image_uri)
 							(EContentEditor *editor);
 
-	void		(*on_cell_dialog_open)		(EContentEditor *editor);
-
-	void		(*on_cell_dialog_close)		(EContentEditor *editor);
-
 	void		(*cell_set_v_align)		(EContentEditor *editor,
 							 const gchar *value,
 							 EContentEditorScope scope);
@@ -404,52 +402,59 @@ struct _EContentEditorInterface {
 	void		(*table_set_background_color)	(EContentEditor *editor,
 							 const GdkRGBA *value);
 
-	gboolean	(*on_table_dialog_open)		(EContentEditor *editor);
-
-	void		(*on_table_dialog_close)	(EContentEditor *editor);
-
-	void		(*on_spell_check_dialog_open)	(EContentEditor *editor);
-
-	void		(*on_spell_check_dialog_close)	(EContentEditor *editor);
-
 	gchar *		(*spell_check_next_word)	(EContentEditor *editor,
 							 const gchar *word);
 
 	gchar *		(*spell_check_prev_word)	(EContentEditor *editor,
 							 const gchar *word);
 
-	void		(*on_replace_dialog_open)	(EContentEditor *editor);
-
-	void		(*on_replace_dialog_close)	(EContentEditor *editor);
-
-	void		(*on_find_dialog_open)		(EContentEditor *editor);
-
-	void		(*on_find_dialog_close)		(EContentEditor *editor);
-
 	/* Signals */
 	void		(*load_finished)		(EContentEditor *editor);
 	gboolean	(*paste_clipboard)		(EContentEditor *editor);
 	gboolean	(*paste_primary_clipboard)	(EContentEditor *editor);
-	gboolean	(*context_menu_requested)	(EContentEditor *editor,
+	void		(*context_menu_requested)	(EContentEditor *editor,
 							 EContentEditorNodeFlags flags,
+							 const gchar *caret_word,
 							 GdkEvent *event);
 	void		(*find_done)			(EContentEditor *editor,
 							 guint match_count);
 	void		(*replace_all_done)		(EContentEditor *editor,
 							 guint replaced_count);
 	void		(*drop_handled)			(EContentEditor *editor);
+	void		(*content_changed)		(EContentEditor *editor);
+	CamelMimePart *	(*ref_mime_part)		(EContentEditor *editor,
+							 const gchar *uri);
+
+	void		(*delete_h_rule)		(EContentEditor *editor);
+	void		(*delete_image)			(EContentEditor *editor);
+
+	gboolean	(*supports_mode)		(EContentEditor *editor,
+							 EContentEditorMode mode);
+	void		(*grab_focus)			(EContentEditor *editor);
+	gboolean	(*is_focus)			(EContentEditor *editor);
+	const gchar *	(*get_hover_uri)		(EContentEditor *editor);
+	void		(*get_caret_client_rect)	(EContentEditor *editor,
+							 GdkRectangle *out_rect);
+
+	/* padding for future expansion */
+	gpointer reserved[15];
 };
 
 /* Properties */
 
 ESpellChecker *	e_content_editor_ref_spell_checker
 						(EContentEditor *editor);
+gboolean	e_content_editor_supports_mode	(EContentEditor *editor,
+						 EContentEditorMode mode);
+void		e_content_editor_grab_focus	(EContentEditor *editor);
+gboolean	e_content_editor_is_focus	(EContentEditor *editor);
+gboolean	e_content_editor_is_malfunction	(EContentEditor *editor);
 gboolean	e_content_editor_can_cut	(EContentEditor *editor);
 gboolean	e_content_editor_can_copy	(EContentEditor *editor);
 gboolean	e_content_editor_can_paste	(EContentEditor *editor);
 gboolean	e_content_editor_can_undo	(EContentEditor *editor);
 gboolean	e_content_editor_can_redo	(EContentEditor *editor);
-gboolean	e_content_editor_is_indented	(EContentEditor *editor);
+gint		e_content_editor_indent_level	(EContentEditor *editor);
 gboolean	e_content_editor_get_spell_check_enabled
 						(EContentEditor *editor);
 void		e_content_editor_set_spell_check_enabled
@@ -461,9 +466,6 @@ void		e_content_editor_set_editable	(EContentEditor *editor,
 gboolean	e_content_editor_get_changed	(EContentEditor *editor);
 void		e_content_editor_set_changed	(EContentEditor *editor,
 						 gboolean changed);
-gboolean	e_content_editor_get_html_mode	(EContentEditor *editor);
-void		e_content_editor_set_html_mode	(EContentEditor *editor,
-						 gboolean html_mode);
 void		e_content_editor_set_alignment	(EContentEditor *editor,
 						 EContentEditorAlignment value);
 EContentEditorAlignment
@@ -494,9 +496,6 @@ gboolean	e_content_editor_is_bold	(EContentEditor *editor);
 void		e_content_editor_set_italic	(EContentEditor *editor,
 						 gboolean italic);
 gboolean	e_content_editor_is_italic	(EContentEditor *editor);
-void		e_content_editor_set_monospaced	(EContentEditor *editor,
-						 gboolean monospaced);
-gboolean	e_content_editor_is_monospaced (EContentEditor *editor);
 void		e_content_editor_set_strikethrough
 						(EContentEditor *editor,
 						 gboolean strikethrough);
@@ -513,6 +512,21 @@ gboolean	e_content_editor_is_superscript
 void		e_content_editor_set_underline	(EContentEditor *editor,
 						 gboolean underline);
 gboolean	e_content_editor_is_underline	(EContentEditor *editor);
+void		e_content_editor_set_start_bottom
+						(EContentEditor *editor,
+						 EThreeState value);
+EThreeState	e_content_editor_get_start_bottom
+						(EContentEditor *editor);
+void		e_content_editor_set_top_signature
+						(EContentEditor *editor,
+						 EThreeState value);
+EThreeState	e_content_editor_get_top_signature
+						(EContentEditor *editor);
+void		e_content_editor_set_visually_wrap_long_lines
+						(EContentEditor *editor,
+						 gboolean value);
+gboolean	e_content_editor_get_visually_wrap_long_lines
+						(EContentEditor *editor);
 
 /* Methods */
 void		e_content_editor_initialize	(EContentEditor *content_editor,
@@ -525,18 +539,51 @@ void		e_content_editor_insert_content	(EContentEditor *editor,
 						 const gchar *content,
 						 EContentEditorInsertContentFlags flags);
 
-gchar *		e_content_editor_get_content	(EContentEditor *editor,
-						 EContentEditorGetContentFlags flags,
+void		e_content_editor_get_content	(EContentEditor *editor,
+						 guint32 flags, /* bit-or of EContentEditorGetContentFlags */
 						 const gchar *inline_images_from_domain,
-						 GSList **inline_images_parts /* newly created CamelMimePart * */);
-
-void            e_content_editor_insert_image_from_mime_part
+						 GCancellable *cancellable,
+						 GAsyncReadyCallback callback,
+						 gpointer user_data);
+EContentEditorContentHash *
+		e_content_editor_get_content_finish
 						(EContentEditor *editor,
-						 CamelMimePart *part);
+						 GAsyncResult *result,
+						 GError **error);
+EContentEditorContentHash *
+		e_content_editor_util_new_content_hash
+						(void);
+void		e_content_editor_util_free_content_hash
+						(EContentEditorContentHash *content_hash);
+void		e_content_editor_util_put_content_data
+						(EContentEditorContentHash *content_hash,
+						 EContentEditorGetContentFlags flag,
+						 const gchar *data);
+void		e_content_editor_util_take_content_data
+						(EContentEditorContentHash *content_hash,
+						 EContentEditorGetContentFlags flag,
+						 gpointer data,
+						 GDestroyNotify destroy_data);
+void		e_content_editor_util_take_content_data_images
+						(EContentEditorContentHash *content_hash,
+						 GSList *image_parts); /* CamelMimePart * */
+gpointer	e_content_editor_util_get_content_data
+						(EContentEditorContentHash *content_hash,
+						 EContentEditorGetContentFlags flag);
+gpointer	e_content_editor_util_steal_content_data
+						(EContentEditorContentHash *content_hash,
+						 EContentEditorGetContentFlags flag,
+						 GDestroyNotify *out_destroy_data);
+CamelMimePart *	e_content_editor_util_create_data_mimepart
+						(const gchar *uri,
+						 const gchar *cid,
+						 gboolean as_inline,
+						 const gchar *prefer_filename,
+						 const gchar *prefer_mime_type,
+						 GCancellable *cancellable);
 
 void		e_content_editor_insert_image	(EContentEditor *editor,
 						 const gchar *uri);
-
 void		e_content_editor_insert_emoticon
 						(EContentEditor *editor,
 						 EEmoticon *emoticon);
@@ -568,9 +615,6 @@ void		e_content_editor_set_spell_checking_languages
 
 void		e_content_editor_select_all	(EContentEditor *editor);
 
-gchar *		e_content_editor_get_selected_text
-						(EContentEditor *editor);
-
 gchar *		e_content_editor_get_caret_word	(EContentEditor *editor);
 
 void		e_content_editor_replace_caret_word
@@ -583,10 +627,6 @@ void		e_content_editor_selection_indent
 void		e_content_editor_selection_unindent
 						(EContentEditor *editor);
 
-void		e_content_editor_selection_create_link
-						(EContentEditor *editor,
-						 const gchar *uri);
-
 void		e_content_editor_selection_unlink
 						(EContentEditor *editor);
 
@@ -609,21 +649,19 @@ void		e_content_editor_selection_restore
 
 void		e_content_editor_selection_wrap	(EContentEditor *editor);
 
-guint		e_content_editor_get_caret_position
-						(EContentEditor *editor);
-
-guint		e_content_editor_get_caret_offset
-						(EContentEditor *editor);
-
 gchar *		e_content_editor_get_current_signature_uid
 						(EContentEditor *editor);
 
 gboolean	e_content_editor_is_ready	(EContentEditor *editor);
+GError *	e_content_editor_dup_last_error	(EContentEditor *editor);
+void		e_content_editor_take_last_error(EContentEditor *editor,
+						 GError *error);
 
 gchar *		e_content_editor_insert_signature
 						(EContentEditor *editor,
 						 const gchar *content,
-						 gboolean is_html,
+						 EContentEditorMode editor_mode,
+						 gboolean can_reposition_caret,
 						 const gchar *signature_id,
 						 gboolean *set_signature_from_message,
 						 gboolean *check_if_signature_is_changed,
@@ -650,11 +688,10 @@ void		e_content_editor_insert_row_above
 void		e_content_editor_insert_row_below
 						(EContentEditor *editor);
 
-gboolean	e_content_editor_on_h_rule_dialog_open
-						(EContentEditor *editor);
-
-void		e_content_editor_on_h_rule_dialog_close
-						(EContentEditor *editor);
+void		e_content_editor_on_dialog_open	(EContentEditor *editor,
+						 const gchar *name);
+void		e_content_editor_on_dialog_close(EContentEditor *editor,
+						 const gchar *name);
 
 void		e_content_editor_h_rule_set_align
 						(EContentEditor *editor,
@@ -686,12 +723,6 @@ void		e_content_editor_h_rule_set_no_shade
 gboolean	e_content_editor_h_rule_get_no_shade
 						(EContentEditor *editor);
 
-void		e_content_editor_on_image_dialog_open
-						(EContentEditor *editor);
-
-void		e_content_editor_on_image_dialog_close
-						(EContentEditor *editor);
-
 void		e_content_editor_image_set_src	(EContentEditor *editor,
 						 const gchar *value);
 
@@ -762,32 +793,22 @@ void		e_content_editor_image_set_height_follow
 						(EContentEditor *editor,
 						 gboolean value);
 
-void		e_content_editor_on_link_dialog_open
-						(EContentEditor *editor);
-
-void		e_content_editor_on_link_dialog_close
-						(EContentEditor *editor);
-
-void		e_content_editor_link_get_values
+void		e_content_editor_link_get_properties
 						(EContentEditor *editor,
-						 gchar **href,
-						 gchar **text);
+						 gchar **out_href,
+						 gchar **out_text,
+						 gchar **out_name);
 
-void		e_content_editor_link_set_values
+void		e_content_editor_link_set_properties
 						(EContentEditor *editor,
 						 const gchar *href,
-						 const gchar *text);
+						 const gchar *text,
+						 const gchar *name);
 
 void		e_content_editor_page_set_text_color
 						(EContentEditor *editor,
 						 const GdkRGBA *value);
 
-void		e_content_editor_on_page_dialog_open
-						(EContentEditor *editor);
-
-void		e_content_editor_on_page_dialog_close
-						(EContentEditor *editor);
-
 void		e_content_editor_page_get_text_color
 						(EContentEditor *editor,
 						 GdkRGBA *value);
@@ -815,6 +836,12 @@ void		e_content_editor_page_set_visited_link_color
 void		e_content_editor_page_get_visited_link_color
 						(EContentEditor *editor,
 						 GdkRGBA *value);
+void		e_content_editor_page_set_font_name
+						(EContentEditor *editor,
+						 const gchar *value);
+
+const gchar *	e_content_editor_page_get_font_name
+						(EContentEditor *editor);
 
 void		e_content_editor_page_set_background_image_uri
 						(EContentEditor *editor,
@@ -823,12 +850,6 @@ void		e_content_editor_page_set_background_image_uri
 gchar *		e_content_editor_page_get_background_image_uri
 						(EContentEditor *editor);
 
-void		e_content_editor_on_cell_dialog_open
-						(EContentEditor *editor);
-
-void		e_content_editor_on_cell_dialog_close
-						(EContentEditor *editor);
-
 void		e_content_editor_cell_set_v_align
 						(EContentEditor *editor,
 						 const gchar *value,
@@ -962,18 +983,6 @@ void		e_content_editor_table_set_background_color
 						(EContentEditor *editor,
 						 const GdkRGBA *value);
 
-gboolean	e_content_editor_on_table_dialog_open
-						(EContentEditor *editor);
-
-void		e_content_editor_on_table_dialog_close
-						(EContentEditor *editor);
-
-void		e_content_editor_on_spell_check_dialog_open
-						(EContentEditor *editor);
-
-void		e_content_editor_on_spell_check_dialog_close
-						(EContentEditor *editor);
-
 gchar *		e_content_editor_spell_check_next_word
 						(EContentEditor *editor,
 						 const gchar *word);
@@ -986,18 +995,13 @@ void		e_content_editor_spell_check_replace_all
 						(EContentEditor *editor,
 						 const gchar *word,
 						 const gchar *replacement);
+void		e_content_editor_delete_h_rule	(EContentEditor *editor);
+void		e_content_editor_delete_image	(EContentEditor *editor);
 
-void		e_content_editor_on_replace_dialog_open
-						(EContentEditor *editor);
-
-void		e_content_editor_on_replace_dialog_close
-						(EContentEditor *editor);
-
-void		e_content_editor_on_find_dialog_open
-						(EContentEditor *editor);
-
-void		e_content_editor_on_find_dialog_close
-						(EContentEditor *editor);
+const gchar *	e_content_editor_get_hover_uri	(EContentEditor *editor);
+void		e_content_editor_get_caret_client_rect
+						(EContentEditor *editor,
+						 GdkRectangle *out_rect);
 
 /* Signal helpers */
 
@@ -1007,9 +1011,10 @@ gboolean	e_content_editor_emit_paste_clipboard
 						(EContentEditor *editor);
 gboolean	e_content_editor_emit_paste_primary_clipboard
 						(EContentEditor *editor);
-gboolean	e_content_editor_emit_context_menu_requested
+void		e_content_editor_emit_context_menu_requested
 						(EContentEditor *editor,
 						 EContentEditorNodeFlags flags,
+						 const gchar *caret_word,
 						 GdkEvent *event);
 void		e_content_editor_emit_find_done	(EContentEditor *editor,
 						 guint match_count);
@@ -1020,6 +1025,13 @@ void		e_content_editor_emit_drop_handled
 						(EContentEditor *editor);
 void		e_content_editor_emit_content_changed
 						(EContentEditor *editor);
+CamelMimePart *	e_content_editor_emit_ref_mime_part
+						(EContentEditor *editor,
+						 const gchar *uri);
+
+gboolean	e_content_editor_util_three_state_to_bool
+						(EThreeState value,
+						 const gchar *mail_key);
 
 G_END_DECLS
 
